/*
The software in this package is distributed under the GNU General
Public License version 2 (with a special exception described below).

A copy of GNU General Public License (GPL) is included in this distribution,
in the file COPYING.GPL.

As a special exception, if other files instantiate templates or use macros
or inline functions from this file, or you compile this file and link it
with other works to produce a work based on this file, this file
does not by itself cause the resulting work to be covered
by the GNU General Public License.

However the source code for this file must still be made available
in accordance with section (3) of the GNU General Public License.

This exception does not invalidate any other reasons why a work based
on this file might be covered by the GNU General Public License.
*/

#ifndef PSEUDO_IPV6_HEADER_H
#define PSEUDO_IPV6_HEADER_H

//-----------------------------------------------------------------------------
// PseudoIpv6Header
//-----------------------------------------------------------------------------

/**
 * @brief This class represents a pseudo IPv6 header used to compute checksum of
 * transport layer protocols.
 *
 * When TCP runs over IPv6, the method used to compute the checksum is defined
 * in RFC 2460.
 *
 * Pseudo IPv6 header format used by TCP and UDP checksums:
 *
 * @code
 * 0                                             23 24           31
 * +--------------------------------------------------------------+      ---
 * |                                                              |       ^
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                      source IPv6 address                     |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * +--------------------------------------------------------------+       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |   40 bytes
 * |                   destination IPv6 address                   |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * +--------------------------------------------------------------+       |
 * |                                                              |       |
 * |              upper-layer packet length in bytes              |       |
 * |                                                              |       |
 * +-----------------------------------------------+--------------+       |
 * |                                               |              |       |
 * |                     zero                      | next header  |       |
 * |                                               |              |       v
 * +-----------------------------------------------+--------------+      ---
 * @endcode
 */
class PseudoIpv6Header
{
public:
    /// Source IPv6 address
    uint8_t source_address[16];
    /// Destination IPv6 address
    uint8_t destination_address[16];
    /// The length of the upper-layer header and data (e.g. TCP header plus
    /// TCP data).
    uint32_t next_packet_length_in_bytes;
    /// Zero has 24 bits split among three 8 bits, once there is no uint24_t
    uint8_t zero_byte[3];
    /// The header of the next protocol
    uint8_t next_header;

};

#endif // PSEUDO_IPV6_HEADER_H
